import Foundation
import ZIPFoundation

extension Data {
    static func makeRandomData(size: Int) -> Data {
        #if os(macOS) || os(iOS) || os(watchOS) || os(tvOS)
            let bytes = [UInt32](repeating: 0, count: size).map { _ in arc4random() }
        #else
            let bytes = [UInt32](repeating: 0, count: size).map { _ in random() }
        #endif
        return Data(bytes: bytes, count: size)
    }
}

class BenchRunner {
    let workingDirectoryURL = URL(fileURLWithPath: NSTemporaryDirectory()).appendingPathComponent("BenchRunner")

    enum BenchRunnerError: Error {
        case noSuchEntry
    }
    
    enum Tool: String {
        case ZIPFoundation
        case DefaultZIP
    }
    
    enum Operation: String {
        case extract
        case add
        case remove
    }
    
    struct Result: CustomStringConvertible {
        var tool: Tool
        var operation: BenchRunner.Operation
        var size: Int
        var URL: URL
        var duration: UInt64
        var durationInMilliseconds: UInt64 {
            return self.duration/1000000
        }
        
        var description: String {
            let fileSizeFormatter = ByteCountFormatter()
            fileSizeFormatter.countStyle = .binary
            let formattedFileSize = fileSizeFormatter.string(fromByteCount: Int64(self.size))
            return "\(self.tool), \(self.operation), \(formattedFileSize), \(self.URL), \(self.durationInMilliseconds)"
        }
    }
    
    init() {
        do {
            let fileManager = FileManager()
            if fileManager.fileExists(atPath: self.workingDirectoryURL.path) {
                try fileManager.removeItem(at: self.workingDirectoryURL)
            }
            try fileManager.createDirectory(at: self.workingDirectoryURL, withIntermediateDirectories: true, attributes: nil)
        } catch {
            print("Error creaeting benchmark environment.")
        }
    }

    func cleanupAndExit() {
        do {
            try FileManager().removeItem(at: self.workingDirectoryURL)
        } catch {}
        Foundation.exit(0)
    }

    func createFile(of size: Int) throws -> URL {
        let data = Data.makeRandomData(size: size)
        let formatter = ByteCountFormatter()
        let formattedSize = formatter.string(fromByteCount: Int64(size))
        let fileName = "\(formattedSize)_testFile" + ProcessInfo.processInfo.globallyUniqueString
        var fileURL = self.workingDirectoryURL
        fileURL.appendPathComponent(fileName)
        try data.write(to: fileURL)
        return fileURL
    }

    func addEntry(at fileURL: URL, tool: Tool, size: Int) throws -> Result {
        let processInfo = ProcessInfo.processInfo
        let archiveFileName = processInfo.globallyUniqueString
        var archiveURL = self.workingDirectoryURL
        archiveURL.appendPathComponent(archiveFileName)
        archiveURL.appendPathExtension("zip")
        let startTime = DispatchTime.now()
        let archive = Archive(url: archiveURL, accessMode: .create)
        try archive?.addEntry(with: fileURL.lastPathComponent, relativeTo: fileURL.deletingLastPathComponent(), compressionMethod: .deflate)
        let endTime = DispatchTime.now()
        let duration = endTime.uptimeNanoseconds-startTime.uptimeNanoseconds
        return Result(tool: tool, operation: .add, size: size, URL: archiveURL, duration: duration)
    }

    func addProcessEntry(at fileURL: URL, tool: Tool, size: Int) throws -> Result {
        let processInfo = ProcessInfo.processInfo
        let archiveFileName = processInfo.globallyUniqueString
        var archiveURL = self.workingDirectoryURL
        archiveURL.appendPathComponent(archiveFileName)
        archiveURL.appendPathExtension("zip")
        let zipProcess = Process()
        zipProcess.launchPath = "/usr/bin/zip"
        zipProcess.arguments = ["-j", "-q", archiveURL.path, fileURL.path]
        let startTime = DispatchTime.now()
        zipProcess.launch()
        zipProcess.waitUntilExit()
        let endTime = DispatchTime.now()
        let duration = endTime.uptimeNanoseconds-startTime.uptimeNanoseconds
        return Result(tool: tool, operation: .add, size: size, URL: archiveURL, duration: duration)
    }

    func extractEntry(with path: String, from archiveURL: URL, tool: Tool, size: Int) throws -> Result {
        let startTime = DispatchTime.now()
        let archive = Archive(url: archiveURL, accessMode: .read)
        guard let entry = archive?[path] else {
            throw BenchRunnerError.noSuchEntry
        }
        let processInfo = ProcessInfo.processInfo
        let entryFileName = processInfo.globallyUniqueString
        var entryURL = self.workingDirectoryURL
        entryURL.appendPathComponent(entryFileName)
        _ = try archive?.extract(entry, to: entryURL)
        let endTime = DispatchTime.now()
        let duration = endTime.uptimeNanoseconds-startTime.uptimeNanoseconds
        return Result(tool: tool, operation: .extract, size: size, URL: entryURL, duration: duration)
    }

    func extractProcessEntry(with path: String, from archiveURL: URL, tool: Tool, size: Int) throws -> Result {
        let processInfo = ProcessInfo.processInfo
        var outputURL = self.workingDirectoryURL
        let outputPathName = processInfo.globallyUniqueString
        outputURL.appendPathComponent(outputPathName)
        try FileManager().createDirectory(at: outputURL, withIntermediateDirectories: true, attributes: nil)
        let unzipProcess = Process()
        unzipProcess.launchPath = "/usr/bin/unzip"
        unzipProcess.arguments = ["-q", "-j", archiveURL.path, path, "-d", outputURL.path]
        let startTime = DispatchTime.now()
        unzipProcess.launch()
        unzipProcess.waitUntilExit()
        let endTime = DispatchTime.now()
        let duration = endTime.uptimeNanoseconds-startTime.uptimeNanoseconds
        // unzip always extracts into a directory using the original entry name, so we move it
        // to the parent directory so that it lies next to all other test files.
        let extractURL = outputURL.appendingPathComponent(path)
        var finalURL = outputURL.deletingLastPathComponent()
        finalURL.appendPathComponent(processInfo.globallyUniqueString)
        try FileManager().moveItem(at: extractURL, to: finalURL)
        try FileManager().removeItem(at: outputURL)
        return Result(tool: tool, operation: .extract, size: size, URL: finalURL, duration: duration)
    }

    fileprivate func makeFilledArchive(at URL: URL, entrySize: Int) throws -> Archive? {
        let archive = Archive(url: URL, accessMode: .update)
        // We add some dummy entries before performing a remove to
        // prevent that we remove the last entry, which would be a not
        // very representative edge case.
        for _ in 0..<3 {
            let fileURL = try self.createFile(of: entrySize)
            try archive?.addEntry(with: fileURL.lastPathComponent, relativeTo: fileURL.deletingLastPathComponent())
        }
        return archive
    }

    func removeEntry(with path: String, from archiveURL: URL, tool: Tool, size: Int) throws -> Result {
        let archive = try self.makeFilledArchive(at: archiveURL, entrySize: size)
        let startTime = DispatchTime.now()
        guard let entry = archive?[path] else {
            throw BenchRunnerError.noSuchEntry
        }
        try archive?.remove(entry)
        let endTime = DispatchTime.now()
        let duration = endTime.uptimeNanoseconds-startTime.uptimeNanoseconds
        return Result(tool: tool, operation: .remove, size: size, URL: archiveURL, duration: duration)
    }

    func removeProcessEntry(with path: String, from archiveURL: URL, tool: Tool, size: Int) throws -> Result {
        _ = try self.makeFilledArchive(at: archiveURL, entrySize: size)
        let zipProcess = Process()
        zipProcess.launchPath = "/usr/bin/zip"
        zipProcess.arguments = ["-q", "-d", archiveURL.path, path]
        let startTime = DispatchTime.now()
        zipProcess.launch()
        zipProcess.waitUntilExit()
        let endTime = DispatchTime.now()
        let duration = endTime.uptimeNanoseconds-startTime.uptimeNanoseconds
        return Result(tool: tool, operation: .remove, size: size, URL: archiveURL, duration: duration)
    }
    
    func makeAveragedResult(from results: [Result], tool: Tool, operation: BenchRunner.Operation, size: Int) -> Result? {
        let subResults = results.filter { (result) -> Bool in
            return result.tool == tool && result.operation == operation && result.size == size
        }
        let durations = subResults.map { (result) -> UInt64 in
            return result.duration
        }
        guard durations.count > 0 else { return nil }
        let averageDuration = durations.reduce(0, +)/UInt64(durations.count)
        return Result(tool: tool, operation: operation, size: size, URL: URL(fileURLWithPath: "/"), duration: averageDuration)
    }
    
    func iterateEnum<T: Hashable>(_: T.Type) -> AnyIterator<T> {
        var i = 0
        return AnyIterator {
            let next = withUnsafeBytes(of: &i) { $0.load(as: T.self) }
            if next.hashValue != i { return nil }
            i += 1
            return next
        }
    }

    fileprivate func printResult(_ testFileSizes: [Int], _ averagedResults: [BenchRunner.Result]) {
        for operation in self.iterateEnum(Operation.self) {
            print("Size, ", terminator: "")
            for size in testFileSizes {
                if size == testFileSizes.reversed().first {
                    print("\(size)")
                } else {
                    print("\(size), ", terminator: "")
                }
            }
            let tools = self.iterateEnum(Tool.self)
            for tool in tools {
                print("\(tool) \(operation), ", terminator: "")
                for size in testFileSizes {
                    let fullTestFileSize = size*1024*1024
                    guard let filteredResult = averagedResults.filter({ $0.operation == operation && $0.tool == tool && $0.size == fullTestFileSize }).first else {
                        continue
                    }
                    if size == testFileSizes.reversed().first {
                        print(String(filteredResult.durationInMilliseconds))
                    } else {
                        print(String(filteredResult.durationInMilliseconds) + ", ", terminator: "")
                    }
                }
            }
            print("\n", terminator: "")
        }
    }

    func run() {
        var results = [Result]()
        var averagedResults = [Result]()
        let testFileSizes = [1, 2, 4, 8, 16]
        do {
            for testFileSize in testFileSizes {
                for _ in 0..<10 {
                    let fullTestFileSize = testFileSize*1024*1024
                    let testFileURL = try self.createFile(of: fullTestFileSize)
                    let addResult = try self.addEntry(at: testFileURL, tool: .ZIPFoundation, size: fullTestFileSize)
                    results.append(addResult)
                    let addProcessResult = try self.addProcessEntry(at: testFileURL, tool: .DefaultZIP, size: fullTestFileSize)
                    results.append(addProcessResult)
                    let extractResult = try self.extractEntry(with: testFileURL.lastPathComponent, from: addResult.URL, tool: .ZIPFoundation, size: fullTestFileSize)
                    results.append(extractResult)
                    let extractProcessResult = try self.extractProcessEntry(with: testFileURL.lastPathComponent, from: addProcessResult.URL, tool: .DefaultZIP, size: fullTestFileSize)
                    results.append(extractProcessResult)
                    let removeResult = try self.removeEntry(with: testFileURL.lastPathComponent, from: addResult.URL, tool: .ZIPFoundation, size: fullTestFileSize)
                    results.append(removeResult)
                    let removeProcessResult = try self.removeProcessEntry(with: testFileURL.lastPathComponent, from: addProcessResult.URL, tool: .DefaultZIP, size: fullTestFileSize)
                    results.append(removeProcessResult)
                }
            }
            
            for tool in self.iterateEnum(Tool.self) {
                for operation in self.iterateEnum(BenchRunner.Operation.self) {
                    for size in testFileSizes {
                        let fullTestFileSize = size*1024*1024
                        guard let averagedResult = self.makeAveragedResult(from: results, tool: tool,
                                                                           operation: operation, size: fullTestFileSize) else {
                                                                            continue
                        }
                        averagedResults.append(averagedResult)
                    }
                }
            }
            printResult(testFileSizes, averagedResults)
        } catch {
            dump(error)
            self.cleanupAndExit()
        }
        self.cleanupAndExit()
    }
}

let runner = BenchRunner()
runner.run()

